/*
    Fairchild FM24C08U/09U EEPROM MPC85xx I2C Driver.
    
    Copyright (c) 2002, Motorola, Inc
    Xianghua Xiao (x.xiao@motorola.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/init.h>

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,0))
#error "This driver must be run with Kernel 2.4.x and above"
#endif

#ifndef THIS_MODULE
#define THIS_MODULE NULL
#endif

/* Scan from 0xA0 to 0xA8 */
static unsigned short normal_i2c[] = { 0x51, I2C_CLIENT_END };
static unsigned short normal_i2c_range[] = { 0x51, 0x51, I2C_CLIENT_END };

/* Insmod parameters */
I2C_CLIENT_INSMOD;

static int checksum = 0;
MODULE_PARM(checksum, "i");
MODULE_PARM_DESC(checksum,
		 "Only accept eeproms whose checksum is correct");


/* Many constants specified below */
#define DEBUG

/* EEPROM registers */
#define EEPROM_REG_CHECKSUM 0x3f

/* EEPROM memory types: */
#define ONE_K		1
#define TWO_K		2
#define FOUR_K		3
#define EIGHT_K		4
#define SIXTEEN_K	5

/* Conversions */
/* Size of EEPROM in bytes */
#define EEPROM_SIZE 128

/* Each client has this additional data */
struct eeprom_data {
	struct semaphore update_lock;
	char valid;		/* !=0 if following fields are valid */
	unsigned long last_updated;	/* In jiffies */

	u8 data[EEPROM_SIZE];	/* Register values */
#if 0
	int memtype;
#endif
};

#ifdef MODULE
extern int init_module(void);
extern int cleanup_module(void);
#endif				/* MODULE */

#ifdef MODULE
static
#else
extern
#endif
int __init i2c_eeprom_init(void);
static int __init eeprom_cleanup(void);

static int eeprom_attach_adapter(struct i2c_adapter *adapter);
static int eeprom_detect(struct i2c_adapter *adapter, int address,
			 unsigned short flags, int kind);
static int eeprom_detach_client(struct i2c_client *client);
static int eeprom_command(struct i2c_client *client, unsigned int cmd,
			  void *arg);

static void eeprom_inc_use(struct i2c_client *client);
static void eeprom_dec_use(struct i2c_client *client);

#if 0
static int eeprom_write_value(struct i2c_client *client, u8 reg,
			      u8 value);
#endif

static void eeprom_update_client(struct i2c_client *client);


/* This is the driver that will be inserted */
static struct i2c_driver eeprom_driver = {
	/* name */ "SPD EEPROM",
	/* id */ I2C_DRIVERID_EEPROM,
	/* flags */ I2C_DF_NOTIFY,
	/* attach_adapter */ &eeprom_attach_adapter,
	/* detach_client */ &eeprom_detach_client,
	/* command */ &eeprom_command,
	/* inc_use */ &eeprom_inc_use,
	/* dec_use */ &eeprom_dec_use
};

/* Used by init/cleanup */
static int __initdata eeprom_initialized = 0;

static int eeprom_id = 0;

int eeprom_attach_adapter(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter, &addr_data, eeprom_detect);
}

/* This function is called by i2c_detect */
int eeprom_detect(struct i2c_adapter *adapter, int address,
		  unsigned short flags, int kind)
{
	int i, cs;
	struct i2c_client *new_client;
	struct eeprom_data *data;
	int err = 0;
	const char *client_name;

	/* Make sure we aren't probing the ISA bus!! This is just a safety check
	   at this moment; i2c_detect really won't call us. */

	printk("detecting clients...\n"); /*XXX*/
	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA))
		    goto ERROR0;

	/* OK. For now, we presume we have a valid client. We now create the
	   client structure, even though we cannot fill it completely yet.
	   But it allows us to access eeprom_{read,write}_value. */
	if (!(new_client = kmalloc(sizeof(struct i2c_client) +
				   sizeof(struct eeprom_data),
				   GFP_KERNEL))) {
		err = -ENOMEM;
		goto ERROR0;
	}

	data = (struct eeprom_data *) (new_client + 1);
	new_client->addr = address;
	new_client->data = data;
	new_client->adapter = adapter;
	new_client->driver = &eeprom_driver;
	new_client->flags = 0;

	/* Now, we do the remaining detection. It is not there, unless you force
	   the checksum to work out. */
	if (checksum) {
		/* prevent 24RF08 corruption */
		i2c_smbus_write_quick(new_client, 0);
		cs = 0;
		for (i = 0; i <= 0x3e; i++)
			cs += i2c_smbus_read_byte_data(new_client, i);
		cs &= 0xff;
		if (i2c_smbus_read_byte_data
		    (new_client, EEPROM_REG_CHECKSUM) != cs)
			goto ERROR1;
	}

#if 1
	/* XXX Determine the chip type - only one kind supported! */
	if (kind <= 0) {
		client_name = "EEPROM chip";
	} else {
#ifdef DEBUG
		printk("eeprom.o: Internal error: unknown kind (%d)?!?",
		       kind);
#endif
		goto ERROR1;
	}
#endif

	/* Fill in the remaining client fields and put it into the global list */
	strcpy(new_client->name, client_name);

	new_client->id = eeprom_id++;
	data->valid = 0;
	init_MUTEX(&data->update_lock);

	/* Tell the I2C layer a new client has arrived */
	if ((err = i2c_attach_client(new_client)))
		goto ERROR3;

	printk("added clients successfully...\n");
	return 0;

/* OK, this is not exactly good programming practice, usually. But it is
   very code-efficient in this case. */

	i2c_detach_client(new_client);
      ERROR3:
      ERROR1:
	kfree(new_client);
      ERROR0:
	return err;
}

int eeprom_detach_client(struct i2c_client *client)
{
	int err;

	if ((err = i2c_detach_client(client))) {
		printk
		    ("eeprom.o: Client deregistration failed, client not detached.\n");
		return err;
	}

	kfree(client);

	return 0;
}


/* No commands defined yet */
int eeprom_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	return 0;
}

void eeprom_inc_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void eeprom_dec_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

#if 0
/* No writes yet (PAE) */
int eeprom_write_value(struct i2c_client *client, u8 reg, u8 value)
{
	return i2c_smbus_write_byte_data(client, reg, value);
}
#endif

void eeprom_update_client(struct i2c_client *client)
{
	struct eeprom_data *data = client->data;
	int i;

	down(&data->update_lock);

	if ((jiffies - data->last_updated > 300 * HZ) |
	    (jiffies < data->last_updated) || !data->valid) {

#ifdef DEBUG
		printk("Starting eeprom update\n");
#endif

		if (i2c_smbus_write_byte(client, 0)) {
#ifdef DEBUG
			printk("eeprom read start has failed!\n");
#endif
		}
		for (i = 0; i < EEPROM_SIZE; i++) {
			data->data[i] = (u8) i2c_smbus_read_byte(client);
		}

		data->last_updated = jiffies;
		data->valid = 1;
	}

	up(&data->update_lock);
}

int __init i2c_eeprom_init(void)
{
	int res;

	printk("i2c-eeprom.o version 0.01\n");
	eeprom_initialized = 0;
	if ((res = i2c_add_driver(&eeprom_driver))) {
		printk
		    ("eeprom.o: Driver registration failed, module not inserted.\n");
		eeprom_cleanup();
		return res;
	}
	eeprom_initialized++;
	return 0;
}

int __init eeprom_cleanup(void)
{
	int res;

	if (eeprom_initialized >= 1) {
		if ((res = i2c_del_driver(&eeprom_driver))) {
			printk
			    ("eeprom.o: Driver deregistration failed, module not removed.\n");
			return res;
		}
	} else
		eeprom_initialized--;

	return 0;
}

EXPORT_NO_SYMBOLS;

#ifdef MODULE

MODULE_AUTHOR("Xianghua Xiao (x.xiao@motorola.com)");
MODULE_DESCRIPTION("SPD EEPROM driver");

int init_module(void)
{
	return i2c_eeprom_init();
}

int cleanup_module(void)
{
	return eeprom_cleanup();
}

#endif /* MODULE */
