/*
 *
 *  85xx_link.c, version 1.0
 *  Handles list management for basic chaining mode
 *
 *  Based on hhnet_dma from MontaVista
 *
 *  This software may be used and distributed according to the terms of
 *  the GNU Public License, Version 2, incorporated herein by reference.
 *
 *  Contact:  <kumar.gala@motorola.com>
 *
 */
#include "85xx_dma_priv.h"
#include "85xx_dma.h"

/* return a public descriptor from the free list of link elements */
int
mpc85xx_dma_alloc_desc(struct dma_pub **pub, void *dma_ch)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	unsigned long flags;

	Debug((PRINT_PREFIX "mpc85xx_dma_alloc_desc\n"));

	if ((pub == NULL) || (ch == NULL)) {
		Debug((PRINT_PREFIX "mpc85xx_dma_alloc_desc - INVAL\n"));
		return -EINVAL;
	}

	spin_lock_irqsave(&ch->lock, flags);
	if (list_empty(&ch->free_link)) {
		spin_unlock_irqrestore(&ch->lock, flags);
		Debug((PRINT_PREFIX "mpc85xx_dma_alloc_desc - NOMEM\n"));
		return -ENOMEM;
	} else {
		int i = 0;
		*pub = list_entry(ch->free_link.next, struct dma_pub, list);
		list_del(&((*pub)->list));
		(*pub)->src = 0;
		(*pub)->dst = 0;
		(*pub)->size = 0;
		(*pub)->direction = 0;
		(*pub)->callback = 0;
		(*pub)->status = 0;
		for (i = 0; i < 32; i++)
			(*pub)->usr[i] = 0;
	}
	spin_unlock_irqrestore(&ch->lock, flags);

	return 0;
}

/* Return a descriptor that was allocated and never used to the
 * list of free link elements */
int
mpc85xx_dma_free_desc(struct dma_pub *pub, void *dma_ch)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	unsigned int flags;

	Debug((PRINT_PREFIX "mpc85xx_dma_free_desc\n"));

	if ((pub == NULL) || (ch == NULL)) {
		Debug((PRINT_PREFIX "mpc85xx_dma_free_desc - INVAL\n"));
		return -EINVAL;
	}

	spin_lock_irqsave(&ch->lock, flags);
	list_add(&pub->list, &ch->free_link);
	spin_unlock_irqrestore(&ch->lock, flags);

	return 0;
}

/* Queue a public descriptor for xfer */
int
mpc85xx_dma_queue_desc(struct dma_pub *pub, void *dma_ch)
{
	LIST_HEAD(head);

	if (pub == NULL)
		return -EINVAL;
	list_add(&pub->list, &head);
	return mpc85xx_dma_queue_list(&head, dma_ch);
}

/* Allocate a list of public descriptors, pulled from the free
 * list of link elements */
int
mpc85xx_dma_alloc_list(unsigned int count, struct list_head *head, void *dma_ch)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	LIST_HEAD(tmp);
	unsigned long flags;

	Debug((PRINT_PREFIX "mpc85xx_dma_alloc_list\n"));

	if ((ch == NULL) || (head == NULL) || (count == 0))
		return -EINVAL;

	spin_lock_irqsave(&ch->lock, flags);

	/*
	 * allocate descriptors until we satisfy the request or run out of
	 * available descriptors
	 */
	while (count && !list_empty(&ch->free_link)) {
		int i = 0;
		struct dma_pub *pub;
		struct list_head *desc = ch->free_link.next;
		list_del(desc);

		pub = list_entry(desc, struct dma_pub, list);
		pub->src = 0;
		pub->dst = 0;
		pub->size = 0;
		pub->direction = 0;
		pub->callback = 0;
		pub->status = 0;
		for (i = 0; i < 32; i++)
			pub->usr[i] = 0;

		list_add(desc, &tmp);
		count--;
	}

	/*
	 * if the request was not completely satisfied, release the ones we
	 * reserved and return an error indicator
	 */
	if (count) {
		list_splice(&tmp, &ch->free_link);
		spin_unlock_irqrestore(&ch->lock, flags);
		return -ENOMEM;
	}

	spin_unlock_irqrestore(&ch->lock, flags);

	/* move our list to the callers list and return success */
	list_splice(&tmp, head->prev);

	return 0;
}

/* return a list of link elements to the free list of link elements
 * internal version which does all the work, and assumes head, dma_ch are valid */
void
__mpc85xx_dma_free_link(struct list_head *head, void *dma_ch)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	struct list_head *cur;
	struct dma_link_desc *p_linkdesc;
	unsigned long flags;

	Debug((PRINT_PREFIX "__mpc85xx_dma_free_link\n"));

	spin_lock_irqsave(&ch->lock, flags);

	list_for_each(cur, head) {
		p_linkdesc = list_entry(cur, struct dma_link_desc, pub.list);
		p_linkdesc->next_link &= ~DMA_LINK_DESC_EOLND;
	}

	list_splice(head, &ch->free_link);
	spin_unlock_irqrestore(&ch->lock, flags);
}

/* return a list of link elements to the free list of link elements */
int
mpc85xx_dma_free_list(struct list_head *head, void *dma_ch)
{

	Debug((PRINT_PREFIX "mpc85xx_dma_free_list\n"));

	if ((dma_ch == NULL) || (head == NULL))
		return -EINVAL;

	__mpc85xx_dma_free_link(head, dma_ch);

	return 0;
}

/* Used to append one list of link elements to an existing list of link elements */
void
mpc85xx_dma_append_list(struct list_head *list, struct list_head *head)
{
	Debug((PRINT_PREFIX "mpc85xx_dma_append_list\n"));

	if (list_empty(list))
		return;

	if (!list_empty(head)) {
		struct dma_link_desc *end;
		struct dma_link_desc *qlast;
		u32 temp;

		end = list_entry(list->prev, struct dma_link_desc, pub.list);
		qlast = list_entry(head->prev, struct dma_link_desc, pub.list);

		temp = end->next_link;
		end->next_link = qlast->next_link;
		qlast->next_link = temp;
	}

	list_splice(list, head->prev);
}

/* Bottom half handling for simple list of link elements */
void
mpc85xx_dma_bh(void *data)
{
	struct dma_channel *ch = (struct dma_channel *) data;
	struct list_head head;
	struct list_head *cur;
	struct dma_link_desc *p_linkdesc;
	unsigned int flags;
	int stat = 0;

	Debug((PRINT_PREFIX "mpc85xx_dma_bh\n"));

	spin_lock_irqsave(&ch->lock, flags);

	if (list_empty(&ch->complete)) {
		spin_unlock_irqrestore(&ch->lock, flags);
		Debug((PRINT_PREFIX "Complete list empty\n"));
		return;
	}

	list_add(&head, &ch->complete);
	list_del_init(&ch->complete);

	spin_unlock_irqrestore(&ch->lock, flags);

	list_for_each(cur, &head) {
		p_linkdesc = list_entry(cur, struct dma_link_desc, pub.list);

		Debug((PRINT_PREFIX "processing 0x%08x\n",
		       (unsigned int) p_linkdesc));

		if (p_linkdesc->next_link & DMA_LINK_DESC_EOLND) {
			Debug((PRINT_PREFIX "EOLND Set\n"));
			if (p_linkdesc->pub.status & DMA_ERROR)
				stat = -EIO;
			else
				stat = 0;

		}
		p_linkdesc->pub.status = stat;

		if (p_linkdesc->pub.callback)
			p_linkdesc->pub.callback((void *) &p_linkdesc->pub);
	}

	__mpc85xx_dma_free_link(&head, (void *) ch);
}

int
mpc85xx_dma_build_desc(struct dma_link_desc *pdesc, struct dma_link_desc *prev,
		       struct dma_channel *ch)
{
	struct dma_pub *pub = &pdesc->pub;

	Debug((PRINT_PREFIX "mpc85xx_dma_build_desc\n"));

	pdesc->src_attr = (DMA_SATR_SREADTTYPE_SNOOP | DMA_SATR_SSME);
	pdesc->src_addr = virt_to_bus(pub->src);
	pdesc->dst_attr = (DMA_DATR_DWRITETTYPE_SNOOP | DMA_DATR_DSME);
	pdesc->dst_addr = virt_to_bus(pub->dst);
	prev->next_link = pdesc->paddr | DMA_CLNDAR_GLEN;
	pdesc->count = pub->size;

	return 0;
}

EXPORT_SYMBOL(mpc85xx_dma_alloc_desc);
EXPORT_SYMBOL(mpc85xx_dma_free_desc);
EXPORT_SYMBOL(mpc85xx_dma_queue_desc);
EXPORT_SYMBOL(mpc85xx_dma_alloc_list);
EXPORT_SYMBOL(__mpc85xx_dma_free_link);
EXPORT_SYMBOL(mpc85xx_dma_free_list);
EXPORT_SYMBOL(mpc85xx_dma_append_list);
EXPORT_SYMBOL(mpc85xx_dma_bh);
EXPORT_SYMBOL(mpc85xx_dma_build_desc);
