/*
 *
 *    Copyright (c) 1999 Grant Erickson <grant@lcse.umn.edu>
 *
 *    Module name: ppc440gx_pic.c
 *
 *    Description:
 *      Interrupt controller driver for PowerPC 4xx-based processors.
 *
 *      NOTE: Totally hacked up version to handle 440GX's cascade
 *            scheme.  Direct from IBM.
 */

/*
 * The PowerPC 403 cores' Asynchronous Interrupt Controller (AIC) has
 * 32 possible interrupts, a majority of which are not implemented on
 * all cores. There are six configurable, external interrupt pins and
 * there are eight internal interrupts for the on-chip serial port
 * (SPU), DMA controller, and JTAG controller.
 *
 * The PowerPC 405/440 cores' Universal Interrupt Controller (UIC) has
 * 32 possible interrupts as well.  Depending on the core and SoC
 * implementation, a portion of the interrrupts are used for on-chip
 * peripherals and a portion of the interrupts are available to be
 * configured for external devices generating interrupts.
 *
 * The PowerNP and 440GP (and most likely future implementations) have
 * cascaded UICs.
 * 11/13/02 rcblach added to the file tree for the 440gx
 * 11/22/02 rcblach added in code to insure that bits 30 and 31 of uic0 are NOT
 *                  enabled.
 */

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/stddef.h>

#include <asm/processor.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/ppc4xx_pic.h>

/* Global Variables */
struct hw_interrupt_type *ppc4xx_pic;
/*
 * We define 4xxIRQ_InitSenses table thusly:
 * bit 0x1: sense, 1 for edge and 0 for level.
 * bit 0x2: polarity, 0 for negative, 1 for positive.
 */
unsigned int ibm4xxPIC_NumInitSenses __initdata = 0;
unsigned char *ibm4xxPIC_InitSenses __initdata = NULL;

/* Six of one, half dozen of the other....#ifdefs, separate files,
 * other tricks.....
 *
 * There are basically two types of interrupt controllers, the 403 AIC
 * and the "others" with UIC.  I just kept them both here separated
 * with #ifdefs, but it seems to change depending upon how supporting
 * files (like ppc4xx.h) change.		-- Dan.  
 */


//#define UIC_DEBUG


static void
ppc440gx_uic_enable(unsigned int irq)
{
	int bit, word;
	irq_desc_t *desc = irq_desc + irq;

	bit = irq & 0x1f;
	word = irq >> 5;

#ifdef UIC_DEBUG
	printk("ppc440gx_uic_enable - irq %d word %d bit 0x%x\n", irq, word, bit);
#endif
	ppc_cached_irq_mask[word] |= 1 << (31 - bit);
	switch (word) {
	case 0:
		mtdcr(DCRN_UIC_ER(UIC0), ppc_cached_irq_mask[word]);
		if ((mfdcr(DCRN_UIC_TR(UIC0)) & (1 << (31 - bit))) == 0)
			desc->status |= IRQ_LEVEL;
		else
		/* lets hope this works since in linux/irq.h
		 * there is no define for EDGE and it's assumed
		 * once you set status to LEVEL you would not 
		 * want to change it - Armin
		 */
		desc->status = desc->status & ~IRQ_LEVEL;
		break;
	case 1:
		mtdcr(DCRN_UIC_ER(UIC1), ppc_cached_irq_mask[word]);
		if ((mfdcr(DCRN_UIC_TR(UIC1)) & (1 << (31 - bit))) == 0)
			desc->status |= IRQ_LEVEL;
		else
		/* lets hope this works since in linux/irq.h
		 * there is no define for EDGE and it's assumed
		 * once you set status to LEVEL you would not 
		 * want to change it - Armin
		 */
		desc->status = desc->status & ~IRQ_LEVEL;
	
	break;

	case 2:
		mtdcr(DCRN_UIC_ER(UIC2), ppc_cached_irq_mask[word]);
		if ((mfdcr(DCRN_UIC_TR(UIC2)) & (1 << (31 - bit))) == 0)
			desc->status |= IRQ_LEVEL;
		else
		/* lets hope this works since in linux/irq.h
		 * there is no define for EDGE and it's assumed
		 * once you set status to LEVEL you would not 
		 * want to change it - Armin
		 */
		desc->status = desc->status & ~IRQ_LEVEL;

	}

}

static void
ppc440gx_uic_disable(unsigned int irq)
{
	int bit, word;

	bit = irq & 0x1f;
	word = irq >> 5;
#ifdef UIC_DEBUG
	printk("ppc440gx_uic_disable - irq %d word %d bit 0x%x\n", irq, word,
	       bit);
#endif
	ppc_cached_irq_mask[word] &= ~(1 << (31 - bit));
	switch (word) {
	case 0:
		mtdcr(DCRN_UIC_ER(UIC0), ppc_cached_irq_mask[word]);
		break;
	case 1:
		mtdcr(DCRN_UIC_ER(UIC1), ppc_cached_irq_mask[word]);
		break;
	case 2:
		mtdcr(DCRN_UIC_ER(UIC2), ppc_cached_irq_mask[word]);
		break;
	}
}

static void
ppc440gx_uic_disable_and_ack(unsigned int irq)
{
	int bit, word;

	bit = irq & 0x1f;
	word = irq >> 5;

#ifdef UIC_DEBUG
	printk("ppc440gx_uic_disable_and_ack - irq %d word %d bit 0x%x\n", irq,
	       word, bit);
#endif
	ppc_cached_irq_mask[word] &= ~(1 << (31 - bit));
	switch (word) {
	case 0:
		mtdcr(DCRN_UIC_ER(UIC0), ppc_cached_irq_mask[word]);
		mtdcr(DCRN_UIC_SR(UIC0), (1 << (31 - bit)));
		/* ACK cascaded interrupt in UIC0 */
		mtdcr(DCRN_UIC_SR(UICB), UIC0NC); /* ack bit */ 
		/* 1/19/03 rcblach  I was acking uic0 instead of the base */
		break;
	case 1:
		   
		mtdcr(DCRN_UIC_ER(UIC1), ppc_cached_irq_mask[word]);
		mtdcr(DCRN_UIC_SR(UIC1), (1 << (31 - bit)));
		/* ACK cascaded interrupt in UIC0 */
		mtdcr(DCRN_UIC_SR(UICB), UIC1NC);
		break;
	case 2:
		mtdcr(DCRN_UIC_ER(UIC2), ppc_cached_irq_mask[word]);
		mtdcr(DCRN_UIC_SR(UIC2), (1 << (31 - bit)));
		/* ACK cascaded interrupt in UIC0 */
		mtdcr(DCRN_UIC_SR(UICB), UIC2NC);
		break;

	}
}

static void
ppc440gx_uic_end(unsigned int irq)
{
	int bit, word;
	unsigned int tr_bits;

	bit = irq & 0x1f;
	word = irq >> 5;

#ifdef UIC_DEBUG
	printk("ppc440gx_uic_end - irq %d word %d bit 0x%x\n", irq, word, bit);
#endif

	switch (word) {
	case 0:
		tr_bits = mfdcr(DCRN_UIC_TR(UIC0));
		break;
	case 1:
		tr_bits = mfdcr(DCRN_UIC_TR(UIC1));
		break;
	case 2:
		tr_bits=mfdcr(DCRN_UIC_TR(UIC2));
		break;
	}

	if ((tr_bits & (1 << (31 - bit))) == 0) {
		/* level trigger */
		switch (word) {
		case 0:
			mtdcr(DCRN_UIC_SR(UIC0), 1 << (31 - bit));
			mtdcr(DCRN_UIC_SR(UICB),  UIC0NC);
			break;
		case 1:
			mtdcr(DCRN_UIC_SR(UIC1), 1 << (31 - bit));
			/* ACK cascaded interrupt in UIC0 */
			mtdcr(DCRN_UIC_SR(UICB),  UIC1NC);
			break;
		case 2:
			mtdcr(DCRN_UIC_SR(UIC2), 1 << (31 - bit));
			/* ACK cascaded interrupt in UIC0 */
			mtdcr(DCRN_UIC_SR(UICB),  UIC2NC);
			break;

		}
	}

	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
		ppc_cached_irq_mask[word] |= 1 << (31 - bit);
		switch (word) {
		case 0:
			mtdcr(DCRN_UIC_ER(UIC0), ppc_cached_irq_mask[word]);
			break;
		case 1:
			mtdcr(DCRN_UIC_ER(UIC1), ppc_cached_irq_mask[word]);
			break;
		case 2:
			mtdcr(DCRN_UIC_ER(UIC2), ppc_cached_irq_mask[word]);
			break;

		}
	}
}

static struct hw_interrupt_type ppc440gx_uic = {
#if (NR_UICS == 1)
	"IBM UIC",
#else
	"IBM UIC Cascade",
#endif
	NULL,
	NULL,
	ppc440gx_uic_enable,
	ppc440gx_uic_disable,
	ppc440gx_uic_disable_and_ack,
	ppc440gx_uic_end,
	0
};
volatile void debug_cde(unsigned long a)
{;}
;
int
ppc440gx_pic_get_irq(struct pt_regs *regs)
{
	int irq, cas_irq;
	unsigned long bits;
	cas_irq = 0;
	/*
	 * Only report the status of those interrupts that are actually
	 * enabled.
	 * The 440gx has 4 uic's with the base uic being the cascade.  
	 */

	bits = mfdcr(DCRN_UIC_MSR(UICB));

	if ((bits & 0x40000000)==0x40000000)  /* 1/19/02 this was wrong, the non critical interrups are 4 1 4 */
   	{
		bits = mfdcr(DCRN_UIC_MSR(UIC0));
	       	
		irq = ( ffs(bits));
		irq = 32-irq;
			
	} 
	/* 
	 * the second Interrupt controller goes from 32 to 63 
	 */
	else if ((bits & 0x10000000)==0x10000000)  
	{
	
		bits = mfdcr(DCRN_UIC_MSR(UIC1));
		cas_irq = 32-  ffs(bits); 	/* the macro ffs returns the bit position of the leftmost 1 as counted from*/
									/* from the right starting with 1.  Therefor 0x80000000 returns 32 and  */
									/* 0x40000000 returns 31, and 0x00001000 returns 13						*/
		irq = 32 + cas_irq;
	}
	/* 
	 * the second Interrupt controller goes from 64 to 95 
	 */
	else if ((bits & 0x04000000)==0x04000000)  
	{
	
		bits = mfdcr(DCRN_UIC_MSR(UIC2));
		cas_irq = 32- ffs(bits); 	/* the macro ffs returns the bit position of the leftmost 1 as counted from*/
									/* from the right starting with 1.  Therefor 0x80000000 returns 32 and  */
									/* 0x40000000 returns 31, and 0x00001000 returns 13						*/
		irq = 64 + cas_irq;
	}
	else
	{ 
		debug_cde(irq);
		irq=-1;
	}
		

	if (irq == (NR_UIC_IRQS * NR_UICS))
		irq = -1;

#ifdef UIC_DEBUG
	printk("ppc440gx_pic_get_irq - irq %d bit 0x%x\n", irq, bits);
#endif

	return (irq);
}

void __init
ppc4xx_extpic_init(void)
{
	/* set polarity 
	 * 1 = default/pos/rising  , 0= neg/falling internal
	 * 1 = neg/falling , 0= pos/rising external
	 * Sense
	 * 0 = default level internal
	 * 0 = level, 1 = edge external
	 */

	unsigned int sense, irq;
	int bit, word;
	unsigned long ppc_cached_sense_mask[NR_MASK_WORDS];
	unsigned long ppc_cached_pol_mask[NR_MASK_WORDS];
	ppc_cached_sense_mask[0] = 0;
	ppc_cached_sense_mask[1] = 0;
	ppc_cached_sense_mask[2] = 0;
	ppc_cached_pol_mask[0] = 0;
	ppc_cached_pol_mask[1] = 0;
	ppc_cached_pol_mask[2] = 0;

	for (irq = 0; irq < NR_IRQS; irq++) 
	{

		bit = irq & 0x1f;
		word = irq >> 5;

		sense =
		    (irq <
		     ibm4xxPIC_NumInitSenses) ? ibm4xxPIC_InitSenses[irq] : 3;
#ifdef PPC4xx_PIC_DEBUG
		printk("PPC4xx_picext %d word:%x bit:%x sense:%x", irq, word,
		       bit, sense);
#endif
		ppc_cached_sense_mask[word] |=
		    (sense & IBM4xx_IRQ_SNS_MASK) << (31 - bit);
		ppc_cached_pol_mask[word] |=
		    ((sense & IBM4xx_IRQ_POL_MASK) >> 1) << (31 - bit);
		switch (word) 
		{
		case 0:
#ifdef PPC4xx_PIC_DEBUG
			printk("Pol %x ", mfdcr(DCRN_UIC_PR(UIC0)));
			printk("Level %x\n", mfdcr(DCRN_UIC_TR(UIC0)));
#endif
			/* polarity  setting */
			mtdcr(DCRN_UIC_PR(UIC0), ppc_cached_pol_mask[word]);

			/* Level setting */
			mtdcr(DCRN_UIC_TR(UIC0), ppc_cached_sense_mask[word]);

			break;
		case 1:
#ifdef PPC4xx_PIC_DEBUG
			printk("Pol %x ", mfdcr(DCRN_UIC_PR(UIC1)));
			printk("Level %x\n", mfdcr(DCRN_UIC_TR(UIC1)));
#endif
			/* polarity  setting */
			mtdcr(DCRN_UIC_PR(UIC1), ppc_cached_pol_mask[word]);

			/* Level setting */
			mtdcr(DCRN_UIC_TR(UIC1), ppc_cached_sense_mask[word]);

			break;

		case 2:
#ifdef PPC4xx_PIC_DEBUG
			printk("Pol %x ", mfdcr(DCRN_UIC_PR(UIC1)));
			printk("Level %x\n", mfdcr(DCRN_UIC_TR(UIC1)));
#endif
			/* polarity  setting */
			mtdcr(DCRN_UIC_PR(UIC2), ppc_cached_pol_mask[word]);

			/* Level setting */
			mtdcr(DCRN_UIC_TR(UIC2), ppc_cached_sense_mask[word]);

			break;
		}
	}

	{
		/* rcblach 11/25/02 this code is to insure that the bits
		 *                  30 and 31  UIC0 bits enable bits are off */
		unsigned long uicbits;
		/* turn off the bits in the enable register */
		uicbits= mfdcr(DCRN_UIC_ER(UIC0)) & 0xfffffffc0;  /* make sure the last two bits are off */
		mtdcr(DCRN_UIC_ER(UIC0),uicbits);
		/* turn off the bits in the critical register */
		uicbits= mfdcr(DCRN_UIC_CR(UIC0)) & 0xfffffffc0;  /* make sure the last two bits are off */
		mtdcr(DCRN_UIC_CR(UIC0),uicbits);

	}

}
void __init
ppc440gx_pic_init(void)
{

	/*
	 * Disable all external interrupts until they are
	 * explicity requested.
	 */
	ppc_cached_irq_mask[0] = 0;
	ppc_cached_irq_mask[1] = 0;
	ppc_cached_irq_mask[2] = 0;

	/* enable the base uic */
	/*TURN ON THE BASE UIC FOR Normal interrupts */
	mtdcr(DCRN_UIC_ER(UICB),UIC0NC|UIC1NC|UIC2NC);
	mtdcr(DCRN_UIC_CR(UICB),0); /*No critical interrupts */
	
	mtdcr(DCRN_UIC_ER(UIC0), ppc_cached_irq_mask[0]);
	mtdcr(DCRN_UIC_CR(UIC0), 0);
	
	mtdcr(DCRN_UIC_ER(UIC1), ppc_cached_irq_mask[1]);
	mtdcr(DCRN_UIC_CR(UIC1), 0);

	mtdcr(DCRN_UIC_ER(UIC2), ppc_cached_irq_mask[2]);
	mtdcr(DCRN_UIC_CR(UIC2), 0);
	
	if (ibm4xxPIC_InitSenses != NULL)
		ppc4xx_extpic_init();

	/* Clear any pending interrupts */
	mtdcr(DCRN_UIC_SR(UIC2), 0xffffffff);
	mtdcr(DCRN_UIC_SR(UIC1), 0xffffffff);
	mtdcr(DCRN_UIC_SR(UIC0), 0xffffffff);
	mtdcr(DCRN_UIC_SR(UICB), 0xffffffff);
	ppc4xx_pic = &ppc440gx_uic;
	ppc_md.get_irq = ppc440gx_pic_get_irq;

}
